package nl.dancingbear.tunit;


import java.net.UnknownHostException;
import java.util.LinkedList;

import toolbus.AbstractTool;
import aterm.ATerm;
import aterm.ATermFactory;
import aterm.ATermList;

/**
 * The ToolStub class is a stub for testing the data flow
 * of a toolbus process. It can be used to connect to a
 * toolbus process and test the communication of events,
 * ack-events, evals, values and do's to and from the process.
 *
 * Events and values can be sent to the connected process,
 * using the sendEvent() and sendValue() functions.
 * The stub can be ordered to wait for incoming ack-events,
 * evals and do's, using the expectAckEvent(), expectEval()
 * and expectDo() functions. These functions throw an exception
 * when a different term is received than expected.
 *
 * @author Bas Basten & Anton Lycklama a Nijeholt
 * @date 21-02-2007
 */
public class ToolStub extends AbstractTool {
    private String name = "";
    private LinkedList<ATerm> queue = new LinkedList<ATerm>();

    /**
     * Exception used when another term is received from the process
     * than expected.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 21-02-2007
     */
    public class UnexpectedTermException extends Exception {
		private static final long serialVersionUID = -201087455600111545L;

		/**
         * Creator.
         *
         * @author Bas Basten & Anton Lycklama a Nijeholt
         * @param toolName The tool that raised the exception.
         * @param term The term that was expected but not received.
         * @date 21-02-2007
         */
        public UnexpectedTermException(String toolName, ATerm term) {
            super("Tool: " + toolName + " --- Term: " + term.toString());
        }
    }

    /**
     * Exception used when the timeout is exceeded in the expect... functions.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 21-02-2007
     */
    public class TimeoutExceededException extends Exception {
		private static final long serialVersionUID = 2617491934066709577L;

		/**
         * Creator.
         *
         * @author Bas Basten & Anton Lycklama a Nijeholt
         * @param toolName The tool that raised the exception.
         * @param term The term that was expected but never received.
         * @param timeout The exceeded time in milliseconds.
         * @date 21-02-2007
         */
        public TimeoutExceededException(String toolName, ATerm term, long timeout) {
            super("Tool: " + toolName + " --- Timeout: " + String.valueOf(timeout) + " --- Term: " + term.toString());
        }
    }

    /**
     * Constructor.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param factory AtermFactory to create ATerms with. Should be the same for all tools
     * @param name The toolname for communication with the Toolbus. This should match the name of the tool used in the process to connect to.
     * @date 21-02-2007
     */
    public ToolStub(ATermFactory factory, String name) {
        super(factory);
        this.name = name;
    }

    /**
     * Returns the name of the StubTool.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @return The name of this tool.
     * @date 21-02-2007
     */
    public String getName() {
        return this.name;
    }

    /**
     * Initialize the tool with the toolbus connection parameters.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param args The toolbus connection parameters.
     * @exception UnknownHostException
     * @date 21-02-2007
     */
    public void init(String[] args) throws UnknownHostException {
        // Add the name of the tool to its connection parameters
        String[] toolNameArgs = { "-TB_TOOL_NAME", this.name };
        String[] newArgs = ArrayUtil.concatArrays(args, toolNameArgs);
        super.init(newArgs);
    }

    /**
     * checkInputSignature inherited from the AbstractTool. Not used.
     * TBD: maybe use anonymous class?
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @date 21-02-2007
     */
    public void checkInputSignature(ATermList sigs) {
        // No implementation
    }

    /**
     * Send a Value to the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param value The value to send.
     * @date 21-02-2007
     */
    public void sendValue(ATerm value) {
        ATerm data = this.factory.make("snd-value(<term>)", value);
        this.sendTerm(data);
    }

    /**
     * Send an event to the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param event The event to send.
     * @date 21-02-2007
     */
    public void sendEvent(ATerm event) {
        ATerm data = this.factory.make("snd-event(<term>)", event);
        this.sendTerm(data);
    }

    /**
     * Generic function to send a term to the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param term Term to send.
     * @date 21-02-2007
     */
    public void sendTerm(ATerm term) {
        try {
            System.out.println("Tool: " + this.name + " --- Sent: " + term.toString());
            super.sendTerm(term);
        }
        catch (Exception ex) {
            System.out.println(StackTraceUtil.getStackTrace(ex));
        }
    }

    /**
     * This is where the ATerms come in from the toolbus.
     * Inherited from AbstractTool.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param term Incoming term from the toolbus
     * @date 21-02-2007
     */
    public ATerm handler(ATerm term) {
        System.out.println("Tool: " + this.name + " --- Received: " + term.toString());
        this.queue.add(term);
        return null;
    }

    /**
     * Wait until an ack-event is received from the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param event The event to be acknowledged.
     * @param timeout The time in milliseconds to wait until a TimeOutExceededException is thrown.
     * @exception TimeoutExceededException The timeout time is exceeded and the expected ack-event is not received.
     * @exception UnexpectedTermException Another term is received than the one expected.
     * @date 21-02-2007
     */
    public void expectAckEvent(ATerm event, long timeout) throws UnexpectedTermException, TimeoutExceededException {
        ATerm expectedTerm = this.factory.make("rec-ack-event(<term>)", event);
        this.waitForATerm(expectedTerm, timeout);
    }

    /**
     * Wait until a do message is received from the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param doMessage The do message that is expected.
     * @param timeout The time in milliseconds to wait until a TimeOutExceededException is thrown.
     * @exception TimeoutExceededException The timeout time is exceeded and the expected do message is not received.
     * @exception UnexpectedTermException Another term is received than the one expected.
     * @date 21-02-2007
     */
    public void expectDo(ATerm doMessage, long timeout) throws UnexpectedTermException, TimeoutExceededException {
        ATerm expectedTerm = this.factory.make("rec-do(<term>)", doMessage);
        this.waitForATerm(expectedTerm, timeout);
    }

    /**
     * Wait until an eval message is received from the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param eval The eval that is expected.
     * @param timeout The time in milliseconds to wait until a TimeOutExceededException is thrown.
     * @exception TimeoutExceededException The timeout time is exceeded and the expected eval is not received.
     * @exception UnexpectedTermException Another term is received than the one expected.
     * @date 21-02-2007
     */
    public void expectEval(ATerm eval, long timeout) throws UnexpectedTermException, TimeoutExceededException {
        ATerm expectedTerm = this.factory.make("rec-eval(<term>)", eval);
        this.waitForATerm(expectedTerm, timeout);
    }

    /**
     * Generic function to wait until an expected term is
     * received from the connected process.
     *
     * @author Bas Basten & Anton Lycklama a Nijeholt
     * @param expectedTerm The term that is expected.
     * @param timeout The time in milliseconds to wait until a TimeOutExceededException is thrown.
     * @exception TimeoutExceededException The timeout time is exceeded and the expected term is not received.
     * @exception UnexpectedTermException Another term is received than the one expected.
     * @date 21-02-2007
     */
    private void waitForATerm(ATerm expectedTerm, long timeout) throws UnexpectedTermException, TimeoutExceededException {
        long startTime = System.currentTimeMillis();

        // Wait for a term from the toolbus until the timeout exceeds.
        while (this.queue.size() == 0) {
            try {
                Thread.sleep(100);
            }
            catch (Exception ex) {
                // No implementation
            }

            if (System.currentTimeMillis() >= (startTime + timeout)) {
                throw new TimeoutExceededException(this.name, expectedTerm, timeout);
            }
        }

        // Test if the received term equals the expected one.
        ATerm receivedTerm = this.queue.remove();
        if (!expectedTerm.isEqual(receivedTerm)) {
            throw new UnexpectedTermException(this.name, receivedTerm);
        }
    }
}
